!	--------------------------------------------------------------------------------
!	flagsit.h
!	versione 0.3 2004-08-13
! 	Copyright (C) 2004  Alessandro Schillaci
!
! 	flagsit.h is free software; you can redistribute it and/or
! 	modify it under the terms of the GNU General Public License
! 	as published by the Free Software Foundation; either version 2
! 	of the License, or (at your option) any later version.
!
! 	flagsit.h is distributed in the hope that it will be useful,
! 	but WITHOUT ANY WARRANTY; without even the implied warranty of
! 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! 	GNU General Public License for more details.
!
! 	You should have received a copy of the GNU General Public License
! 	along with this program; if not, write to the Free Software
! 	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
! ===============================================================================


!	--------------------------------------------------------------------------------
!	Questa piccola libreria pu essere utilizzata per:
!
!	1) La gestione di "Chapter" (Capitoli) del gioco.
!	2) La gestione di "flag" che indicano se una certa azione
!	    stata oppure no portata a termine
!	3) Visualizzare dei messaggi di aiuto, se il giocatore "ristagna"
!	   troppo a lungo all'interno di un "Capitolo".
!	--------------------------------------------------------------------------------
!	Per Utilizzare la libreria:
!
!	1) Definire la Constante DEBUG_FLAGSIT; (se si vuole il Debug)
!	2) Includere il file di libreria, Include "flagsit";
!	3) Nella funzione Initialise, far partire il daemon "ChapterTurnCounter"
!	   - con StartDaemon(ChapterTurnCounter); -
!	4) Definire una nuove funzione come questa:
!
!		[ printMessages;
!			setMessage(1,3,"Messaggio di aiuto per il 3 turno del capitolo 1");
!			setMessage(2,4,"Messaggio di aiuto per il 4 turno del capitolo 2");
!			setMessage(3,5,"Messaggio di aiuto per il 5 turno del capitolo 3");
!
!			setMessageFlag(1,20,"Forse dovresti ~Aprire il cassetto~",1);
!		];
!
!	  Questa funzione si occupa di visualizzare i messaggi ad un certo turno
!	  di un certo capitolo. Per Esempio, con :
!			setMessage(1,3,"Messaggio di aiuto per il 3 turno del capitolo 1");
!	  Nel Capitolo 1, turno 3, verr visualizzata la stringa.
!
!	  Questo pu essere molto utile per visuazlizzare dei suggerimenti al
!	  giocatore, quando esso si trovi per troppo tempo in una fase (capitolo) del
!	  gioco. Mano a mano che i turni all'interno dello stesso capitolo aumentano
!	  significa che il giocatore potrebbe essere "bloccato" in quella fase di gioco.
!	--------------------------------------------------------------------------------
!	Note:
!
!	Sono disponibili le seguenti funzioni:
!
!		getChapter()			= Restituisce il numero del capitolo corrente
!		setChapter(x)			= Imposta il Capitolo corrente
!		setFlag(x,onoff)		= Imposta il flag "x" al valore onnoff (0,1)
!		setMessageFlag(x,y,z,w)	= Stampa il messaggio z, al turno y del capitolo x se il flag w non  settato
!		flagOn()				= Ritorna true se il flag==1,false se il flag != 1
!	--------------------------------------------------------------------------------


Constant LUNGH_MAX 50;
Global chapter = 0;

Array flag -> LUNGH_MAX;

! Contatore dei turni che si azzera quando viene variata la variabile chapter
Object ChapterTurnCounter
  with daemon [;
    	++self.turn_active;
    	#ifdef DEBUG_FLAGSIT;
	  	print "^FLAGSIT DEBUG: Chapter ", chapter, ", Turno attuale ", self.turn_active;
  	#endif;
  	printMessages();
  ],
  turn_active 0;



! Gestione dei Capitoli dell'avventura
[ setChapter x ;
	chapter = x;
    	#ifdef DEBUG_FLAGSIT;
		print  "^FLAGSIT DEBUG: Impostato Chapter = ", x, "^";
	#endif;
	ChapterTurnCounter.turn_active = 0;
];

[ getChapter ;
	return chapter;
];




! gestione azioni flag: possono servire per cambiare capitolo
[ setFlag x onoff;
	flag ->x = onoff;
	#ifdef DEBUG_FLAGSIT;
		print "^FLAGSIT DEBUG: Ho impostato il flag ", x, " a ",  onoff, "^";
	#endif;
];
[ flagOn x;
  if (flag->x==1) { rtrue; }
  rfalse;
];



! Funzione che stampa il messaggio ad un certo turno all'interno di un certo
! Chapter
! x = il capitolo
! y = il turno del capitolo x
! z = il messaggio di aiuto da stampare
[ setMessage x y z;
	#ifdef DEBUG_FLAGSIT;
		print "^FLAGSIT DEBUG: Condizione da testare CAPITOLO=", x, ", TURNO=", y;
	#endif;
	if ((getChapter()==x)&&(ChapterTurnCounter.turn_active == y)){
		print "^",(string) z;
	}
];

! Funzione che stampa il messaggio ad un certo turno all'interno di un certo
! Chapter se il flag w non  settato
! x = il capitolo
! y = il turno del capitolo x
! z = il messaggio di aiuto da stampare
! w = il flag da controllare
[ setMessageFlag x y z w;
	#ifdef DEBUG_FLAGSIT;
		print "^FLAGSIT DEBUG: Condizione da testare CAPITOLO=", x, ", TURNO=", y," FLAG ",w,"= 1";
		print "^Stato FLAG ",w,"= ",flagOn(w);
	#endif;

	if ((getChapter()==x)&&(ChapterTurnCounter.turn_active == y) && ~~flagOn(w)){
		print "^",(string) z;
	}
];



