!   footnote_it.h       An Inform Library Extention for Inform 5.5 or greater
!                       Per produrre note a pie' di pagina numerate in sequenza
!                       L. Ross Raszewski
!                       Traduzione e addattamento all'italiano di P.Lucchesi
!       VERSION 2.0
! New features: *Define Redisplay_Footnotes to 1 before inclusion to prevent
!               the player from seeing the footnote indicator more than once
!               *Define footnote1__tx to the text to be printed before the
!               number and footnote2__tx to the text to be printed after it
!               in order to change the default [footnote #] message
!               (multi-lingual support)
!
! Nuove caratteristiche:
!     *Definite Redisplay_Footnotes a 1 prima dell'inclusione per impedire
!     che il giocatore veda l'indicatore di nota piu' di una volta
!     *Definite footnote1__tx come testo da stampare prima del numero e
!     footnote2__tx come testo da stampare dopo, in modo da cambiare il
!     messaggio di default [nota #] messaggio (supporto multilingua)
!
!
!
!
! Questa libreria, basata su di un esercizio nel manuale, vi permette di
! integrare note a pie' di pagina nel gioco, numerate NELL'ORDINE IN CUI
! APPAIONO. Questo porta dei vantaggi rispettao al semplice, ovvio sistema
! (quello usato in Hitchhiker's Guide): Primo, non si puo' accedere ale note
! che non sono state ancora menzionate. Secondo, le note sono sempre numerate
! in sequenza ascendente, indipendentemente dall'ordine in cui sono chiamate.
! L'esempio nel manuel, ho trovato, qualche volta assegna piu' numeri alla
! stessa nota. Ho evitato questo, in modo che la stessa nota abbia sempre lo
! stesso numero, indipendentemente da come viene chiamata (un po' come
! Achieved(#);)
!
! Includetela DOPO grammar.h
! Prima dell'inclusione definite la costante MAX_FOOTNOTES come numero totale
! delle note nel gioco.
! Per stampare una nota nel testo del gioco, chiamate Note() con il numero
! della nota (il numero e' solo interno, e di solito non e' quello che appare
! a schermo. Esempio: Se due note sono gia' state viste, allora l'istruzione:
!      Print "Tutto il mondo e' un palcoscenico.", (note) 1;
! stampera' "Tutto il mondo e' un palcoscenico [Nota 3]", con [Nota 3]
! stampato in underline. Note() stampa il testo contenuto in Footnote1__TX e
! Foornote2__TX, che di default hanno uno spazio in testa ma non in coda.
!
! La nota viene letta dando il comando NOTA 3 (o LEGGI NOTA 3). NOTA da solo
! istruira' il giocatore sull'uso delle note a pie' di pagina.
!
! Dovete definire la funzione PrintNote(n); prima dell'inclusione. Questa
! funzione attualmente stampa il testo della nota dopo che FootnoteSub ha
! stampato le parole [Footnote #]^ e ha convertito il numero che il giocatore
! ha scrinno nel numero originale nella lista delle note. Una semplice
! funzione PrintNote potrebbe essere:
!       [ PrintNote n;
!         switch(n){
!                 1: "William Shakespeare";
!                 2: "Gli Sloath non hanno gusto";
!                 3: {Style bold; print "Sesame Street"; style roman; " vi \
!                       e' stato offerto da ~K~ e il numero 4.";};
!                 };
!              ];
!
! Una nota puo' fare quello che volete, anche chiamare un'altra nota (l'ho
! provato, pare che funzioni).
! Un attimo di attenzione: L'array footnotes_seen e' un array di bit, cosi'
! penso che non possiate avere piu' di 256 note.
!
!

System_file;
Array footnotes_seen -> MAX_FOOTNOTES;
Global footnote_count;
#IFNDEF Redisplay_Footnotes;
Constant Redisplay_Footnotes 0;
#ENDIF;

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Multi-lingual block.  Define the following 
! contants before inclusion to use alternate 
! Messages
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#IFNDEF Footnote1__TX;
Constant Footnote1__TX "[Nota ";
#ENDIF;
#IFNDEF Footnote2__TX;
Constant Footnote2__TX "]";
#ENDIF;
#IFNDEF Footnote3__TX;
Constant Footnote3__TX "Le note iniziano da 1.";
#ENDIF;
#IFNDEF Footnote4__TX;
Constant Footnote4__TX "La nota [";
#ENDIF;
#IFNDEF Footnote5__TX;
Constant Footnote5__TX "] non  stata menzionata.";
#ENDIF;
#IFNDEF Footnote6__TX;
Constant Footnote6__TX "Per vedere una nota, scrivete ~NOTA #~, dove
                        #  il numero della nota in questione.";
#ENDIF;
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! End of Multi-lingual block
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

[ FNWarnSub;
        print_ret (string) Footnote6__TX;
        ];
[ Note n;
     if (footnotes_seen->n~=0 && Redisplay_Footnotes==1) rfalse;
     if (footnotes_seen->n==0) {
     footnote_count++; footnotes_seen->n=footnote_count;};
     print " ";
     style underline;
     print (string) Footnote1__TX,footnotes_seen->n, (string) Footnote2__TX;
     style roman;
];
[ FootnoteSub n;
     if (noun>footnote_count)
     {   print (string) Footnote4__TX,noun, (string) Footnote5__TX, "^"; rtrue; }
     if (noun==0) print_ret (string) Footnote3__tx;
     for(n=1:(footnotes_seen->n~=noun && n<=(MAX_FOOTNOTES-1)):n++);
     style underline;
     print (string) Footnote1__TX,noun,(string) Footnote2__tx, "^";
     style roman;
     PrintNote(n);
 ];

Verb meta "nota" *                                      ->FNWarn
           * number                                     ->Footnote;
Extend "leggi"
           * "nota" number                              ->Footnote
           * "nota"                                     ->FNWarn
           * "note"                                     ->FNWarn;

#Stub PrintNote 1;
